;;
;; BrainF*ck compiler for SHARP PC-1500 (LH5801)
;; Copyright 2015 Christophe Gottheimer <cgh75015@gmail.com>

;;
;; BrainF_ck compiler is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License version 2 as published
;; by the Free Software Foundation.  Note that I am not granting permission to
;; redistribute or modify the BrainF_ck compiler under the terms of any later
;; version of the General Public License.
;; 
;; This program is distributed in the hope that it will be useful (or at
;; least amusing), but WITHOUT ANY WARRANTY; without even the implied warranty
;; of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
;; Public License for more details.
;;
;; You should have received a copy of the GNU General Public License along with
;; this program (in the file "COPYING"); if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	.CODE
	.COMMENT:	BrainF*ck compiler - Copyright 2015 Christophe Gottheimer <cgh75015@gmail.com>

%00e	.EQU	00
%00p	.EQU	00

	;; Compiler entry point
BF_compiler:
	SCF	;; Compiler has CARRY set
	JR	BF_common

	;; Compiled code entry point
BF_code@call:
	RCF	;; Code has CARRY cleared

BF_common:
	PUSH	DE
	LD	BC,PC

	;; Because compiler is relocatable, compute the offset of
	;; the last RET
	.IF	PASS?	2
%00e	.EQU	BF_code@start
%00e	.EQU	[-'.]%00e
%00p	.EQU	}%00e
%00e	.EQU	[-'%00p]%00e
	.ENDIF

	LDA	*_*BF_code@epilog
	LD	DE,BC
	ADD	DE

	LDA	%00e
	ADD	BC
	LDA	%00p
	ADD	BC

	;; CARRY ? Yes ! Go to the compiler
	JR	C,BF_docompile

BF_code@prolog:
	;; Prolog - Push the epilog return address
	PUSH	DE

	;; Prolog - Clear the screen
	CLS
	;; Clear the buffer area from A$..D$ (&78C0..&78FF) and
	;; A..X (&7900..&79BF)
	LD	D,<[+07]_X
	LD	E,>[+07]_X
	CLA		; &34 ::= LDA N
	STA	L
	DEC	L

BF_zeroloop:
	DEC	DE
	STA	(DE)
	DJC	BF_zeroloop

	;; The current pointer DE is set to A (&7900)
	LD	D,<_A
	LD	E,>_A

	;; Jump to compiled code
	LD	PC,BC

BF_code@epilog:
	POP	DE
	RCF
	RET

BF_docompile:
	;; DE points to the destination code address
	LD	DE,BC

	;; In command ?
	IFC
	JR	Z,BF_comp7865

	;; NO. Retrieve the source address from the stack
	LD	BC,SP
	LDA	06
	ADD	BC
	LDD	(BC)
	STA	L
	LDA	(BC)
	STA	B
	LDA	L
	STA	C

BF_compcrloop:
	;; Fetch the &0D, the end-of-line
	LDI	(BC)
	CPA	&0D
	JR	NZ,BF_compcrloop
	DEC	BC
	JR	BF_comp

	;; YES. The source address is taken from STARTBASIC (&7865..&7866)
BF_comp7865:
	LDS	(65)

BF_comp:
	LD	L,&01
	LD	H,&00
	DEC	BC
	JR	BF_comploop

BF_compldloop:
	;; Load the code. The destination is DE. The source is BC and the
	;; length is A. Check for code overwriting over BASIC
	PUSH	A
	LDI	(BC)
	STI	(DE)
	LDA	D

	;; Does code overwrite the BASIC area ?
	CPA	(7865)
	JR	Z,BF_compmemov
	POP	A
	DEC	A
	JR	NZ,BF_compldloop

BF_comploopPOP:
	POP	BC
	INC	BC

BF_comploop:
	;; L is 0. No more caractere to read on the current line
	DEC	L
	JR	NZ,BF_compdo

	;; BC points on &0D. Increment and load byte pointed by BC
	INC	BC
	LDI	(BC)

	;; &FF ? End of BASIC program
	CPA	FF
	JR	Z,BF_compend

	;; Skip the line number and load the line length pointed by BC
	INC	BC
	LDI	(BC)
	;; Store the length in L
	STA	L
	JR	BF_comploop

BF_compdo:
	;; Fetch instruction
	LDA	(BC)
	PUSH	BC

	;; < ::= Decremente pointer
	CPA	$<
	JR	NZ,BF_compinc

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_<
	ADD	BC
	LDA	01
	JR	BF_compldloop

BF_compinc:
	;; > ::= Incremente pointer
	CPA	$>
	JR	NZ,BF_compsub

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_>
	ADD	BC
	LDA	01
	JR	BF_compldloop

BF_compsub:
	;; - ::= Substract 1 from the pointed byte
	CPA	$-
	JR	NZ,BF_compadd

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_-
	ADD	BC
	LDA	02
	JR	BF_compldloop

BF_compadd:
	;; - ::= Add 1 to the pointed byte
	CPA	$+
	JR	NZ,BF_compwhile

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_+
	ADD	BC
	LDA	02
	JR	BF_compldloop

BF_compwhile:
	;; This is the instruction [. But PC1500 does not
	;; have the [. So use the ( instead
	;; ( ::= Check if pointed byte is zero. If YES,
	;; jump to the instruction after the ). If NO,
	;; execute the following code
	CPA	$:squareroot	;; [ is SQRT sign. Alias it to (
	JR	Z,+04
	CPA	$(
	JR	NZ,BF_compwhileend

	;; The current "loop" address is pushed into our
	;; stack. Use 3 bytes for each (
	LDA	H
	CPA	#207
	JR	C,BF_compstkov
	STA	C
	LD	B,&79

	;; Build the JP <after-)>
	LDA	OPCODE'jp:&mn
	STI	(BC)
	LDA	D
	STI	(BC)
	LDA	E
	STI	(BC)
	LDA	C
	STA	H

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_while
	ADD	BC
	LDA	[-'BF_while]BF_while_end
	JR	BF_compldloop

BF_compwhileend:
	;; This is the instruction ]. But PC1500 does not
	;; have the ]. So use the ) instead
	;; ) ::= Return to the ( instruction
	CPA	$:pi	;; ] is PI sign. Alias it to )
	JR	Z,+04
	CPA	$)
	JR	NZ,BF_compinput

	;; Be sure to 3 bytes pushed into the stack
	LDA	H
	CPA	03
	JR	NC,BF_compstkov

	;; Pop the return-to-( address
	STA	C
	DEC	C
	LD	B,&79
	LDD	(BC)
	PUSH	A
	LDD	(BC)
	PUSH	A
	LDA	C
	STA	H
	POP	BC

	;; Skip the LDA (DE); JR NZ,+01; JP
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	PUSH	DE

	;; Add 3 byte to skip the JP zztt, because we jump
	;; after the )
	INC	DE
	INC	DE
	INC	DE

	;; Patch the JP xxyy to write the address after the
	;; current )
	LDA	D
	STI	(BC)
	LDA	E
	STI	(BC)
	POP	DE

	;; BC points to code to write and A its length
	LDA	H
	STA	C
	LD	B,&79
	LDA	03
	JR	BF_compldloop

BF_compinput:
	;; , ::= Read a character from the keyboard and store it
	;; into the byte pointed
	;; If ENTER is read, put the value 0 (cgh)
	CPA	$:comma
	JR	NZ,BF_compprint

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_comma
	ADD	BC
	LDA	[-'BF_comma]BF_comma_end
	JR	BF_compldloop
BF_compprint:
	;; . ::= Write the byte pointed to the LCD.
	CPA	$.
	JR	NZ,BF_comploopPOP

	;; BC points to code to write and A its length
	LD	BC,PC
	LDA	*_*BF_.
	ADD	BC
	LDA	[-'BF_.]BF_._end
	JR	BF_compldloop

BF_compend:
	;; End of compilation. Just add RET to return to epilog
	LDA	OPCODE'ret
	STI	(DE)
	LDA	D
	;; Does code overwrite the BASIC area ?
	CPA	(7865)
	JR	Z,BF_compmemov

BF_compexit:
	;; Restore the register and returns
	POP	DE
	RCF

	;; That's all folks !
	RET

BF_compmemov:
	LD	H,#10	; Overflow of code over BASIC
	POP	A

BF_comperr:
	POP	DE
	ERRH

BF_compstkov:
	LD	H,#16	; ( ) stack overflow
	JR	BF_comperr

BF_<:
	;; < ::= Decrement the pointer
	DEC	DE

BF_>:
	;; < ::= Increment the pointer
	INC	DE

BF_-:
	;; - ::= Substract 1 from the pointed byte
	ADD	(DE),&FF

BF_+:
	;; - ::= Add 1 to the pointed byte
	ADD	(DE),&01

BF_while:
	;; This is the instruction [. But PC1500 does not
	;; have the [. So use the ( instead
	;; ( ::= Check if pointed byte is zero. If YES,
	;; jump to the instruction after the ). If NO,
	;; execute the following code
	LDA	(DE)
	JR	NZ,+03
	;; The JP xxyy is merged with the endwhile to save
	;; some bytes...
BF_endwhile:
	;; This is the instruction ]. But PC1500 does not
	;; have the ]. So use the ) instead
	;; ) ::= Return to the ( instruction
	JP	E000	; dummy address to patch
BF_endwhile_end:
BF_while_end:


BF_.:
	;; . ::= Write the byte pointed to the LCD.
	LDA	(DE)
	PUSH	DE
	CALL	ED4D
	POP	DE
BF_._end:

BF_comma:
	;; , ::= Read a character from the keyboard and store it
	;; into the byte pointed
	;; If ENTER is read, put the value 0 (cgh)
	PUSH	DE
	CALL	E24A
	POP	DE
	CPA	0F	; If OFF, put -1 (EOF)
	JR	NZ,+02
	LDA	FF
	CPA	0D	; If ENTER, put 0
	JR	NZ,+01
	CLA		; &34 ::= LDA N
	STA	(DE)
BF_comma_end:

BF_code@start:
	;; Compiled code entry. If no code compiled, just return
	RET

	;; That's all folks !
	.END

